package ru.rutoken.Samples.rtJNA;

import com.sun.jna.Native;
import com.sun.jna.ptr.IntByReference;
import ru.rutoken.Pkcs11.*;

class RtPkcs11Library {
    public static Pkcs11 INSTANCE = (Pkcs11) Native.loadLibrary("rtPKCS11ECP", Pkcs11.class);
    RtPkcs11Library() {}
    public static Pkcs11 getInstance() {
        return INSTANCE;
    }
}

//class InternalException extends Exception {
//    InternalException(String text) {
//        super(text);
//    }
//}

public class Init{

    private final static byte[] USER_PIN = {'1', '2', '3', '4', '5', '6', '7', '8'};
    private final static byte[] SO_PIN = {'8', '7', '6', '5', '4', '3', '2', '1'};
    private final static byte[] TOKEN_LABEL = {'M', 'y', ' ', 'R', 'u', 't', 'o', 'k',
                                               'e', 'n', ' ', ' ', ' ', ' ', ' ', ' ',
                                               ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ',
                                               ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ' };
    public static void main(String[] args) {
        int hSession = -1;
        try{
            System.out.println("Example of token initialization using rtPKCS11ECP.dll via JNA");

            System.out.println("Library initialization and getting of function list");
            int rv = RtPkcs11Library.getInstance().C_Initialize(null);
            if (rv != 0)
            {
                throw new Exception("Failed to initialize library, error code:" + Integer.toHexString(rv));
            }

            System.out.println("Getting list of slots with connected tokens");

            IntByReference slotsCount = new IntByReference();
            slotsCount.setValue(5);
            int[] pSlotList = new int[5];

            rv = RtPkcs11Library.getInstance().C_GetSlotList(true, null, slotsCount);
            if (rv != 0)
            {
                throw new Exception("C_GetSlotList failed, error code: " + Integer.toHexString(rv));
            }
            if (slotsCount.getValue() == 0)
            {
                throw new Exception("No Rutoken ECP is available!");
            }

            rv = RtPkcs11Library.getInstance().C_GetSlotList(true, pSlotList, slotsCount);
            if (rv != 0)
            {
                throw new Exception("C_GetSlotList failed, error code: " + Integer.toHexString(rv));
            }

            System.out.println("Token initialization");
            rv = RtPkcs11Library.getInstance().C_InitToken(pSlotList[0],SO_PIN, SO_PIN.length, TOKEN_LABEL);
            if (rv != 0)
            {
                throw new Exception("C_InitToken failed, error code: " + Integer.toHexString(rv));
            }

            System.out.println("Opening of session");
            IntByReference phSession = new IntByReference();
            rv = RtPkcs11Library.getInstance().C_OpenSession(pSlotList[0],
                    Pkcs11Constants.CKF_SERIAL_SESSION | Pkcs11Constants.CKF_RW_SESSION, null, null, phSession);
            if (rv != 0)
            {
                throw new Exception("C_OpenSession failed, error code: " + Integer.toHexString(rv));
            }
            hSession = phSession.getValue();

            System.out.println("Logging in as administrator");
            rv = RtPkcs11Library.getInstance().C_Login(hSession, Pkcs11Constants.CKU_SO, SO_PIN, SO_PIN.length);
            if (rv != 0)
            {
                throw new Exception("C_Login failed, error code: " + Integer.toHexString(rv));
            }

            System.out.println("User PIN initialization");
            rv = RtPkcs11Library.getInstance().C_InitPIN(hSession, USER_PIN, USER_PIN.length);
            if (rv != 0)
            {
                throw new Exception("C_InitPIN failed, error code: " + Integer.toHexString(rv));
            }

            System.out.println("Logging out");
            rv = RtPkcs11Library.getInstance().C_Logout(hSession);
            if (rv != 0)
            {
                throw new Exception("C_Logout failed, error code: " + Integer.toHexString(rv));
            }

            System.out.println("Closing of session");
            rv = RtPkcs11Library.getInstance().C_CloseSession(hSession);
            if (rv != 0)
            {
                throw new Exception("C_CloseSession failed, error code: " + Integer.toHexString(rv));
            }
            hSession = -1;

            System.out.println("Test has been completed successfully.\n");
        }
        catch(Exception e){
            System.out.printf(e.getMessage() + "\nSome error occurred.");
        }
        try {
            RtPkcs11Library.getInstance().C_Logout(hSession);
        }
        catch(Exception e){}
        try {
            RtPkcs11Library.getInstance().C_CloseSession(hSession);
        }
        catch(Exception e){}
        try {
            RtPkcs11Library.getInstance().C_Finalize(null);
        }
        catch(Exception e){}
    }
}
