/*************************************************************************
* Rutoken                                                                *
* Copyright (c) 2003-2015, CJSC Aktiv-Soft. All Rights Reserved.         *
* Подробная информация:  http://www.rutoken.ru                           *
* Загрузка драйверов:    http://www.rutoken.ru/hotline/download/drivers/ *
* Техническая поддержка: http://www.rutoken.ru/hotline/                  *
*------------------------------------------------------------------------*
* Пример работы с Рутокен ЭЦП Flash при помощи библиотеки PKCS#11        *
* на языке C                                                             *
*------------------------------------------------------------------------*
* Использование команд управления памятью Рутокен Flash:                 *
*  - установление соединения с Рутокен в первом доступном слоте;         *
*  - получение объема Flash памяти Рутокен;								 *
*  - выполнение аутентификации с правами Администратора;                 *
*  - создание PIN-кода Локального Пользователя Рутокен;                  *
*  - создание разделов Flash памяти Рутокен;                             *
*  - получение информации о разделах Flash памяти Рутокен;               *
*  - изменение атрибутов разделов Flash памяти Рутокен;                  *
*  - сброс прав доступа Администратора и закрытие соединения с Рутокен.  *
*------------------------------------------------------------------------*
* Данный пример является самодостаточным из серии примеров работы        *
* с библиотекой PKCS#11                                                  *
*************************************************************************/

#include "Common.h"

/************************************************************************
* main()                                                                *
************************************************************************/
int main(int argc, char* argv[])
{
	HMODULE hModule = NULL_PTR;                            // Хэндл загруженной библиотеки PKCS#11
	CK_SESSION_HANDLE hSession = NULL_PTR;                 // Хэндл открытой сессии

	CK_FUNCTION_LIST_PTR pFunctionList = NULL_PTR;         // Указатель на список функций PKCS#11, хранящийся в структуре CK_FUNCTION_LIST
	CK_C_GetFunctionList pfGetFunctionList = NULL_PTR;     // Указатель на функцию C_GetFunctionList

	CK_FUNCTION_LIST_EXTENDED_PTR pFunctionListEx = NULL_PTR;        // Указатель на список функций расширения PKCS#11, хранящийся в структуре CK_FUNCTION_LIST_EXTENDED
	CK_C_EX_GetFunctionListExtended pfGetFunctionListEx = NULL_PTR;  // Указатель на функцию C_EX_GetFunctionListExtended
	CK_TOKEN_INFO_EXTENDED tokenInfoEx;                              // Структура данных типа CK_TOKEN_INFO_EXTENDED с информацией о токене

	CK_SLOT_ID_PTR aSlots = NULL_PTR;                      // Указатель на массив идентификаторов всех доступных слотов
	CK_ULONG ulSlotCount = 0;                              // Количество идентификаторов всех доступных слотов в массиве

	CK_ULONG ulDriveSize = 0;							   // Общий объем носителя

	CK_VOLUME_ID_EXTENDED	VolumeRW = 01;				   // Идентификатор раздела для чтения и записи
	CK_VOLUME_ID_EXTENDED	VolumeRO = 02;				   // Идентификатор раздела только для чтения
	CK_VOLUME_ID_EXTENDED	VolumeHI = 03;				   // Идентификатор раздела CD-ROM
	CK_VOLUME_ID_EXTENDED	VolumeCD = 04;				   // Идентификатор скрытого раздела

	CK_ULONG				VolumeRWSize = 0;              // Размер раздела для чтения и записи
	CK_ULONG				VolumeROSize = 0;              // Размер раздела только для чтения
	CK_ULONG				VolumeCDSize = 0;              // Размер раздела CD-ROM
	CK_ULONG				VolumeHISize = 0;              // Размер раздела скрытого раздела


	CK_ULONG			    CKU_LOCAL_1 = 0x03;			   // Идентификатор локального пользователя
	CK_ULONG			    CKU_LOCAL_2 = 0x1F - 1;        // Идентификатор локального пользователя


	/* DEMO PIN-код локального пользователя Рутокен */
	static CK_UTF8CHAR      LOCAL_PIN[] = { '1', '2', '3', '4', '5', '6', '7', '8', '9', '0' };

	/* Шаблон для разметки разделов */
	CK_VOLUME_FORMAT_INFO_EXTENDED InitParams[] =
	{
		{ VolumeRWSize, ACCESS_MODE_RW, CKU_USER, 0 },
		{ VolumeROSize, ACCESS_MODE_RO, CKU_SO, 0 },
		{ VolumeHISize, ACCESS_MODE_HIDDEN, CKU_LOCAL_1, 0 },
		{ VolumeCDSize, ACCESS_MODE_CD, CKU_LOCAL_2, 0 }
	};
	
	CK_VOLUME_INFO_EXTENDED_PTR   pVolumesInfo = NULL_PTR;  // Указатель на массив структур с информацией о разделах
	CK_ULONG                      ulVolumesInfoSize = 0;    // Размер массива с информацией о разделах

	CK_ULONG	i = 0;                                      // Вспомогательная переменная для счетика циклов
	CK_RV		rv = CKR_OK;                                // Вспомогательная переменная для хранения кода возврата
	CK_RV		rvTemp = CKR_OK;                            // Вспомогательная переменная для хранения кода возврата

	printf("Initialization module...\n");
	while (TRUE)
	{
		/**********************************************************************
		* Шаг 1: Загрузить библиотеку.                                        *
		**********************************************************************/
		printf(" Loading library %s", PKCS11ECP_LIBRARY_NAME);   
		hModule = LoadLibrary(PKCS11ECP_LIBRARY_NAME);
		if (hModule == NULL_PTR)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* Шаг 2: Получить адрес функции запроса структуры с указателями       *
		*        на функции.                                                  *
		**********************************************************************/
		printf(" Getting GetFunctionList function");
		pfGetFunctionList = (CK_C_GetFunctionList)GetProcAddress(hModule,
		                                                         "C_GetFunctionList");
		if (pfGetFunctionList == NULL_PTR)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* Шаг 3: Получение адреса функции запроса структуры с указателями     *
		*        на функции расширения стандарта PKCS#11.                     *
		**********************************************************************/
		printf("Getting GetFunctionListExtended function");
		pfGetFunctionListEx = (CK_C_EX_GetFunctionListExtended)GetProcAddress(hModule,
			"C_EX_GetFunctionListExtended");
		if (pfGetFunctionListEx == NULL_PTR)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* Шаг 4: Получение структуры с указателями на функции                 *
		*        стандарта PKCS#11.                                           *
		**********************************************************************/
		printf("Getting function list");
		rv = pfGetFunctionList(&pFunctionList);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* Шаг 5: Получение структуры с указателями на функции расширения      *
		*        стандарта PKCS#11.                                           *
		**********************************************************************/
		printf("Getting extended function list");
		rv = pfGetFunctionListEx(&pFunctionListEx);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* Шаг 6: Инициализация библиотеки.                                    *
		**********************************************************************/
		printf("Initializing library");
		rv = pFunctionList->C_Initialize(NULL_PTR);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* Шаг 7: Получить количество слотов c подключенными токенами.         *
		**********************************************************************/
		/* Определение количества подключенных слотов */
		printf("Getting number of connected slots");
		rv = pFunctionList->C_GetSlotList(CK_TRUE,
		                                  NULL_PTR,
		                                  &ulSlotCount);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		if (ulSlotCount == 0)
		{
			printf(" No Rutoken is available!\n");
			break;
		}

		aSlots = (CK_SLOT_ID*)malloc(ulSlotCount * sizeof(CK_SLOT_ID));
		memset(aSlots,
		       0,
		       (ulSlotCount * sizeof(CK_SLOT_ID)));

		/**********************************************************************
		* Шаг 8: Получить список слотов c подключенными токенами.             *
		**********************************************************************/
		printf("Getting list of connected slots");
		rv = pFunctionList->C_GetSlotList(CK_TRUE,
		                                  aSlots,
		                                  &ulSlotCount);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		printf(" Slots available: 0x%8.8X\n", (int)ulSlotCount);

		/**********************************************************************
		* Шаг 9: Определить класс токена.                                     *
		**********************************************************************/
		printf("Determining token type");

		/**********************************************************************
		* Получить расширенную информацию о подключенном токене.              *
		**********************************************************************/
		tokenInfoEx.ulSizeofThisStructure = sizeof(tokenInfoEx); 
		rv = pFunctionListEx->C_EX_GetTokenInfoExtended(aSlots[0],
													   &tokenInfoEx);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}

		/**********************************************************************
		* Определить класс токена.                                            *
		**********************************************************************/
	    if (tokenInfoEx.ulTokenClass == TOKEN_CLASS_ECP)
		{
			printf(": Rutoken ECP\n");
		}
		else
		{

			printf(": No Rutoken ECP\n");
			break;
		}

		/**********************************************************************
		* Шаг 10: Создать PIN-код для локального пользователя токена.         *
		**********************************************************************/
		printf("\nSet Local User PIN");
		rv = pFunctionListEx->C_EX_SetLocalPIN( aSlots[0],
												USER_PIN,
												sizeof(USER_PIN),
												LOCAL_PIN,
												sizeof(LOCAL_PIN),
												CKU_LOCAL_1);
		if (rv != CKR_OK)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");


		/**********************************************************************
		* Шаг 11: Получить объем Flash-памяти.                                *
		**********************************************************************/
		printf("\nGet Flash memory size");
		rv = pFunctionListEx->C_EX_GetDriveSize(aSlots[0],
												&ulDriveSize);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		printf(" Memory size: %d Mb\n", (int)ulDriveSize);

		VolumeRWSize = ulDriveSize / 2;
		VolumeROSize = ulDriveSize / 4;
		VolumeHISize = ulDriveSize / 8;
		VolumeCDSize = ulDriveSize - (ulDriveSize / 2) - (ulDriveSize / 4) - (ulDriveSize / 8);

		InitParams[0].ulVolumeSize = VolumeRWSize;
		InitParams[1].ulVolumeSize = VolumeROSize;
		InitParams[2].ulVolumeSize = VolumeHISize;
		InitParams[3].ulVolumeSize = VolumeCDSize;

		/**********************************************************************
		* Шаг 12: Форматировать Flash память.			                      *
		**********************************************************************/
		printf("\nFormatting Flash memory");
		rv = pFunctionListEx->C_EX_FormatDrive( aSlots[0],
												CKU_SO,
												SO_PIN,
												sizeof(SO_PIN),
												InitParams,
												arraysize(InitParams));
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* Шаг 13: Получить информацию о разделах Flash памяти.	              *
		**********************************************************************/
		printf("\nGetting volumes info");
		rv = pFunctionListEx->C_EX_GetVolumesInfo( aSlots[0],
												   NULL_PTR,
												   &ulVolumesInfoSize);


		pVolumesInfo = (CK_VOLUME_INFO_EXTENDED*)malloc(ulVolumesInfoSize * sizeof(CK_VOLUME_INFO_EXTENDED));
		memset(pVolumesInfo,
				0,
				(ulVolumesInfoSize * sizeof(CK_ULONG)));
		rv = pFunctionListEx->C_EX_GetVolumesInfo(aSlots[0],
													pVolumesInfo,
													&ulVolumesInfoSize);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		for (i = 0; i < (unsigned)ulVolumesInfoSize; i++)
		{   
			printf("\nPrinting volume %1.2d info:\n", (int)i+1);
			printf(" Volume id:         %2.2d \n", (int)pVolumesInfo[i].idVolume);
		    printf(" Volume size:       %d Mb \n", (int)pVolumesInfo[i].ulVolumeSize);
			printf(" Access mode:       %2.2d \n", (int)pVolumesInfo[i].accessMode);
			printf(" Volume owner:      %2.2d \n", (int)pVolumesInfo[i].volumeOwner);
			printf(" Flags:             0x%8.8X \n", (unsigned)pVolumesInfo[i].flags);
		}

		/**********************************************************************
		* Шаг 14: Изменить атрибуты разделов Flash памяти.	                  *
		**********************************************************************/
		printf("\nChanging volume attributes");

		rv = pFunctionListEx->C_EX_ChangeVolumeAttributes(aSlots[0],
															CKU_SO,
															SO_PIN,
															sizeof(SO_PIN),
															VolumeRO,
															ACCESS_MODE_RW,
															CK_TRUE);       // Постоянное изменение атрибутов
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");
		 
		break;
	}

	printf("\nFinalizing... \n");


	if (hSession)
	{
		/**********************************************************************
		* Шаг 15: Закрыть все открытые сессии в слоте                         *
		**********************************************************************/
		printf(" C_CloseAllSession");
		rvTemp = pFunctionList->C_CloseAllSessions(aSlots[0]);
		if (rvTemp != CKR_OK)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");
		hSession = NULL_PTR;
	}

	if (pFunctionList)
	{
		/**********************************************************************
		* Шаг 16: Деинициализировать библиотеку                               *
		**********************************************************************/
		printf(" Finalizing library");
		rvTemp = pFunctionList->C_Finalize(NULL_PTR);
		if (rvTemp != CKR_OK)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");
		pFunctionList = NULL_PTR;
	}

	if (hModule)
	{
		/**********************************************************************
		* Шаг 17: Выгрузить библиотеку из памяти                              *
		**********************************************************************/
		printf(" Unloading library");
		if (FreeLibrary(hModule) != TRUE)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");
		hModule = NULL_PTR;
	}

	if (rvTemp == CKR_OK)
		printf("Unloading has been completed successfully.\n\n");
	else
		printf("Unloading failed!\n\n");

	if (pVolumesInfo)
	{
		free(pVolumesInfo);
		pVolumesInfo = NULL_PTR;
	}

	if (aSlots)
	{
		free(aSlots);
		aSlots = NULL_PTR;
	}

	if ((rv != CKR_OK) || (ulSlotCount == 0))
		printf("Some error occurred. Error code: 0x%8.8x. Press Enter to exit.\n", (int)rv);
	else if (rvTemp != CKR_OK)
		printf("Some error occurred. Error code: 0x%8.8x. Press Enter to exit.\n", (int)rvTemp);
	else
		printf("Test has been completed successfully. Press Enter to exit.\n");


	getchar();
	return rv != CKR_OK;
}
