/*************************************************************************
* Rutoken                                                                *
* Copyright (c) 2003-2015, CJSC Aktiv-Soft. All rights reserved.         *
* Подробная информация:  http://www.rutoken.ru                           *
* Загрузка драйверов:    http://www.rutoken.ru/hotline/download/drivers/ *
* Техническая поддержка: http://www.rutoken.ru/hotline/                  *
*------------------------------------------------------------------------*
* Пример работы с Рутокен PINPad при помощи библиотеки PKCS#11           *
* на языке C                                                             *
*------------------------------------------------------------------------*
* Использование команд работы с журналом:                                *
*  - установление соединения с Рутокен PINPad в первом доступном слоте;  *
*  - выполнение аутентификации c правами Пользователя;					 *
*  - получение содержимого журнала Рутокен PINPad;                       *
*  - разбор данных в журнале                                             *
*  - сброс прав доступа Пользователя на Рутокен PINPad и закрытие        *
*    соединения с Рутокен PINPad.                                        *
*------------------------------------------------------------------------*
* Пример самодостаточен                                                  *
*************************************************************************/

#include "Common.h"

typedef enum {
	TLVTJournal = 0x80,
	TLVTHash = 0xAA,
	TLVTSignature = 0xB6,
	TLVTOperationInfo = 0x85,
	TLVTPinPadId = 0x83,
	TLVTTagTableInfo = 0x86
} TLVType;

typedef enum {
	OTSignature = 0x01,
} OperationType;

typedef enum 
{
	RSFSTypeGCHV = 0x01,
	RSFSTypeLCHV = 0x21,
	RSFSTypeSGOST = 0x02,
	RSFSTypeAES = 0x22,
	RSFSType3DES = 0x42,
	RSFSTypeAGOSTPR = 0x03,
	RSFSTypeRSAPR = 0x23,
	RSFSTypeAGOST_512PR = 0x43,
	RSFSTypeAGOSTPU = 0x13,
	RSFSTypeRSAPU = 0x33,
	RSFSTypeAGOST_512PU = 0x53,
	RSFSTypeX509CER = 0x14,
	RSFSTypeACGOST = 0x05,
	RSFSTypeSE = 0x1F
} RSFSubType;

typedef enum
{
	RSFFAllowKeyExch = 0x01,
	RSFFKeyForSM = 0x02,
	RSFFKeySignsJournal = 0x04,
	RSFFKeyImported = 0x08,
	RSFFKeyPINEnter = 0x10,
	RSFFKeyConfirmOp = 0x20
} RSFFlags;

typedef enum
{
	OIDeviceHash = 0x01,
	OIKeyPINEnter = 0x02,
	OIPIN2Entered = 0x10,
	OIOpConfirmed = 0x20
} OperationInfoFlags;

typedef enum
{
	PPPPIN2CashDisabled = 0x01,
	PPPDefPINNotRequested = 0x02,
	PPPChangePINOnScreen = 0x04
} PinPadPolicyFlags;

#pragma pack(push, 1)
typedef struct OperationInfo_ {
	uint8_t operationType;
	uint8_t rsfType;
	uint8_t rsfFlags;
	uint8_t oiFlags;
	uint8_t pppFlags;
	uint8_t reserved;
	uint8_t rsfIDHigh;
	uint8_t rsfIDLow;
	uint32_t ulSignatureCount; 
} OperationInfo;

typedef struct TagTableInfo_ {
	uint16_t tableNumber;
	uint16_t tableEncoding;
	uint8_t tableHash[32];
} TagTableInfo;
#pragma pack(pop)

typedef struct TLVPrinter_ {
	TLVType type;
	void (*print)(const CK_BYTE*);
} TLVPrinter;

uint16_t be2les(uint16_t be) {
	uint16_t le;
	int i = 0;
	for (i = 0; i < sizeof(be); ++i) {
		le = (le << 8) | (be & 0xFF);
		be >>= 8;
	}
	return le;
}

uint32_t be2lei(uint32_t be) {
	uint32_t le;
	int i = 0;
	for (i = 0; i < sizeof(be); ++i) {
		le = (le << 8) | (be & 0xFF);
		be >>= 8;
	}
	return le;
}

const CK_BYTE* FindTlVElement(CK_BYTE ubTag, const CK_BYTE* pFirstTLVElt, CK_ULONG ulLen)
{
	const CK_BYTE* pLastValueByte = pFirstTLVElt + ulLen;
	const CK_BYTE* pTmpTag = pFirstTLVElt;

	while (pTmpTag < pLastValueByte)
	{
		if (*pTmpTag == ubTag)
			return pTmpTag;

		pTmpTag += *(pTmpTag + 1) + 2;
	}

	return 0; // tag is not found
}

const CK_BYTE* GetTlVValue(const CK_BYTE* pTLVElt) {
	return pTLVElt+2;
}

CK_ULONG GetTlVLen(const CK_BYTE* pTLVElt) {
	return *(pTLVElt+1);
}

void printHexBuffer(const CK_BYTE* pBuffer, const CK_ULONG ulLen, const CK_ULONG offset) {
	unsigned int i = 0, j=0;
	for (i = 0; i < ulLen; ++i)
	{
		if(i%8 == 0)
			for(j = 0; j < offset; ++j)
				printf("\t");

		printf("%02X ", pBuffer[i]);

		if ((i + 1) % 8 == 0 || (i + 1) == ulLen)
			printf("\n");
	}
}

void printOperationType(OperationType operationType)
{
	printf("\t\tOperation Type: ");
	switch(operationType)  {
		case OTSignature: printf("Signature"); break;
		default: printf("unknown (%02X)", operationType);
	}
	printf("\n");
}

void printRSFType(RSFSubType rsfType) {
	printf("\t\tRSF Type: ");
	switch(rsfType)  {
		case RSFSTypeGCHV: printf("GCHV (Global PIN)"); break;
		case RSFSTypeLCHV: printf("LCHV (Local PIN)"); break;
		case RSFSTypeSGOST: printf("SGOST (GOST 28147-89 key)"); break;
		case RSFSTypeAES: printf("AES"); break;
		case RSFSType3DES: printf("3DES"); break;
		case RSFSTypeAGOSTPR: printf("AGOST_PR (GOST 34.10-2001 private key)"); break;
		case RSFSTypeRSAPR: printf("RSA_PR (RSA private key)"); break;
		case RSFSTypeAGOST_512PR: printf("AGOST2012_PR (GOST 34.10-2012 private key)"); break;
		case RSFSTypeAGOSTPU: printf("AGOST_PU (GOST 34.10-2001 public key)"); break;
		case RSFSTypeRSAPU: printf("RSA_PU (RSA public key)"); break;
		case RSFSTypeAGOST_512PU: printf("AGOST2012_PU (GOST 34.10-2012 public key)"); break;
		case RSFSTypeX509CER: printf("X509 (X509 certificate)"); break;
		case RSFSTypeACGOST: printf("ACGOST (GOST 28147-89 key)"); break;
		case RSFSTypeSE: printf("SE (SE environment object)"); break;
		default: printf("unknown (%02X)", rsfType);
	}
	printf("\n");
}

void printRSFFlags(RSFFlags rsfFlags) {
	printf("\t\tRSF Flags: \n");
	if (rsfFlags&RSFFAllowKeyExch) printf("\t\t\tKey Exchange Allowed\n");
	if (rsfFlags&RSFFKeyForSM) printf("\t\t\tKey used for SM\n");
	if (rsfFlags&RSFFKeySignsJournal) printf("\t\t\tKey used to sign journal\n");
	if (rsfFlags&RSFFKeyImported) printf("\t\t\tKey was imported\n");
	if (rsfFlags&RSFFKeyPINEnter) printf("\t\t\tKey usage requires PIN2 enter\n");
	if (rsfFlags&RSFFKeyConfirmOp) printf("\t\t\tKey usage requires operation confirmation\n");
	if (!rsfFlags)
		printf("\t\t\tNone\n");
}

void printOperationInfoFlags(OperationInfoFlags oiFlags) {
	printf("\t\tOperation Info Flags: \n");
	if (oiFlags&OIDeviceHash) printf("\t\t\tHash was performed on device\n");
	if (oiFlags&OIKeyPINEnter) printf("\t\t\tKey used requires PIN2 enter\n");
	if (oiFlags&OIPIN2Entered) printf("\t\t\tPIN2 was entered\n");
	if (oiFlags&OIOpConfirmed) printf("\t\t\tOperation was confirmed\n");
	if (!oiFlags)
		printf("\t\t\tNone\n");
}

void printPinPadPolicyFlags(PinPadPolicyFlags pppFlags) {
	printf("\t\tPinPad Policy Flags: \n");
	if (pppFlags&PPPPIN2CashDisabled) printf("\t\t\tPIN2 cache disabled\n");
	if (pppFlags&PPPDefPINNotRequested) printf("\t\t\tDefault PIN2 not requested on PIN2 change\n");
	if (pppFlags&PPPChangePINOnScreen) printf("\t\t\tPIN2 can be change on screen only\n");
	if (!pppFlags)
		printf("\t\t\tNone\n");
}

void printRSFID(CK_BYTE rsfIDHigh, CK_BYTE rsfIDLow) {
	unsigned int rsfID = ((unsigned int)rsfIDHigh<<8) | rsfIDLow;
	printf("\t\tRSF Id: %04X\n", rsfID);
}

void printSignatureCount(uint32_t ulSignatureCount) {
	uint32_t ulSignatureCountLE = 0;
	unsigned int i = 0;
	for (i = 0; i < sizeof(ulSignatureCount); ++i) {
		ulSignatureCountLE = (ulSignatureCountLE << 8) | (ulSignatureCount & 0xFF);
		ulSignatureCount >>= 8;
	}
	printf("\t\tSuccessfull Signature Count: %d\n", ulSignatureCountLE);
}

void printOperationInfo(const CK_BYTE* pTLVElt) {
	const OperationInfo* pOperationInfo = (const OperationInfo*)GetTlVValue(pTLVElt);
	printf("\tOperation Info:\n");
	printOperationType(pOperationInfo->operationType);
	printRSFType(pOperationInfo->rsfType);
	printRSFFlags(pOperationInfo->rsfFlags);
	printOperationInfoFlags(pOperationInfo->oiFlags);
	printPinPadPolicyFlags(pOperationInfo->pppFlags);
	printRSFID(pOperationInfo->rsfIDHigh, pOperationInfo->rsfIDLow);
	printSignatureCount(pOperationInfo->ulSignatureCount);
}

void printTableNumber(uint16_t tableNumber) {
	printf("\t\tTable number: %d\n", be2les(tableNumber));
}

void printTableEncoding(uint16_t tableEncoding) {
	printf("\t\tTable encoding: ");
	switch(be2les(tableEncoding)) {
		case 0: printf("CP-1251"); break;
		case 1: printf("UTF-8"); break;
		default: printf("unknown");
	}
	printf("\n");
}

void printTableHash(const CK_BYTE* pTableHash, CK_ULONG ulLen) {
	printf("\t\tHash:\n");
	printHexBuffer(pTableHash, ulLen, 3);
}

void printTagTableInfo(const CK_BYTE* pTLVElt) {
	const TagTableInfo* pTagTableInfo = (const TagTableInfo*)GetTlVValue(pTLVElt);
	printf("\tTag Table Info:\n");
	printTableNumber(pTagTableInfo->tableNumber);
	printTableEncoding(pTagTableInfo->tableEncoding);
	printTableHash(pTagTableInfo->tableHash, sizeof(pTagTableInfo->tableHash));
}

void printHash(const CK_BYTE* pTLVElt) {
	printf("\tHash:\n");
	printHexBuffer(GetTlVValue(pTLVElt), GetTlVLen(pTLVElt), 2);
}

void printSignature(const CK_BYTE* pTLVElt) {
	printf("\tSignature:\n");
	printHexBuffer(GetTlVValue(pTLVElt), GetTlVLen(pTLVElt), 2);
}

void printPinPadId(const CK_BYTE* pTLVElt) {
	unsigned int i = 0;
	const CK_BYTE* id = GetTlVValue(pTLVElt);
	printf("\tPinPad Id: ");
	for (i = 0; i < GetTlVLen(pTLVElt); ++i)
		printf("%02X", id[i]);
	printf("\n");
}


void printJournal(const CK_BYTE* pBuffer, CK_ULONG ulLength)
{
	const CK_BYTE* pJournal = NULL;
	const CK_BYTE* pJournalBody = NULL;
	CK_ULONG ulJournalLength = 0;
	TLVPrinter journalPrinters[] = {
		{TLVTHash, printHash},
		{TLVTSignature, printSignature},
		{TLVTOperationInfo, printOperationInfo},
		{TLVTPinPadId, printPinPadId},
		{TLVTTagTableInfo, printTagTableInfo}
	};
	unsigned int i = 0;

	printf("PinPad Journal: ");
	pJournal = FindTlVElement(TLVTJournal, pBuffer, ulLength);
	if (!pJournal) {
		printf("not found\n");
		return;
	} else {
		printf("\n");
	}

	pJournalBody = GetTlVValue(pJournal);
	ulJournalLength = GetTlVLen(pJournal);

	for (i = 0; i < arraysize(journalPrinters); ++i) {
		const CK_BYTE* pTLVElt = FindTlVElement(journalPrinters[i].type, pJournalBody, ulJournalLength);
		if (pTLVElt)
			journalPrinters[i].print(pTLVElt);
	}
}


/************************************************************************
* main()                                                                *
************************************************************************/
int main(void)
{
	HMODULE hModule = NULL_PTR;                          // Хэндл загруженной библиотеки PKCS#11
	CK_SESSION_HANDLE hSession = NULL_PTR;               // Хэндл открытой сессии

	CK_FUNCTION_LIST_PTR pFunctionList = NULL_PTR;       // Указатель на список функций PKCS#11, хранящийся в структуре CK_FUNCTION_LIST
	CK_C_GetFunctionList pfGetFunctionList = NULL_PTR;   // Указатель на функцию C_GetFunctionList

	CK_C_EX_GetFunctionListExtended pfGetFunctionListEx = NULL_PTR; // Указатель на функцию C_EX_GetFunctionListExtended
	CK_FUNCTION_LIST_EXTENDED_PTR pFunctionListEx = NULL_PTR;       // Указатель на список функций расширения PKCS#11, хранящийся в структуре CK_FUNCTION_LIST_EXTENDED
	CK_TOKEN_INFO_EXTENDED tokenInfoEx;                             // Структура данных типа CK_TOKEN_INFO_EXTENDED с информацией о токене

	CK_SLOT_ID_PTR aSlots = NULL_PTR;                    // Указатель на массив идентификаторов слотов
	CK_ULONG ulSlotCount = 0;                            // Количество идентификаторов слотов в массиве

	CK_BYTE_PTR pJournal = NULL_PTR;					 // Указатель на значение журнала
	CK_ULONG ulJournalSize = 0;							 // Размер журнала

	CK_BBOOL bIsRutokenPINPad = FALSE;                   // Вспомогательная переменная для хранения признака типа токена
	CK_RV rv = CKR_OK;                                   // Вспомогательная переменная для хранения кода возврата
	CK_RV rvTemp = CKR_OK;                               // Вспомогательная переменная для хранения кода возврата

	while (TRUE)
	{
		/**********************************************************************
		* Шаг 1: Выполнить действия для начала работы с библиотекой PKCS#11.  *
		**********************************************************************/
		printf("Initialization module...\n");
		while (TRUE)
		{
			/**********************************************************************
			* 1.1 Загрузить библиотеку                                            *
			**********************************************************************/
			printf("Loading library %s", PKCS11ECP_LIBRARY_NAME);
			hModule = LoadLibrary(PKCS11ECP_LIBRARY_NAME);
			if (hModule == NULL_PTR)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/************************************************************************
			* 1.2 Получить адрес функции запроса структуры с указателями на функции *
			************************************************************************/
			printf(" Getting GetFunctionList function");
			pfGetFunctionList = (CK_C_GetFunctionList)GetProcAddress(hModule,
			                                                         "C_GetFunctionList");
			if (pfGetFunctionList == NULL_PTR)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/**********************************************************************
			* 1.3 Получить структуру с указателями на функции                     *
			**********************************************************************/
			printf(" Getting function list");
			rv = pfGetFunctionList(&pFunctionList);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/**********************************************************************
			* 1.4 Инициализировать библиотеку                                     *
			**********************************************************************/
			printf(" Initializing library");
			rv = pFunctionList->C_Initialize(NULL_PTR);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/**********************************************************************
			* 1.5 Получить количество слотов c подключенными токенами             *
			**********************************************************************/
			printf(" Getting number of connected slots");
			rv = pFunctionList->C_GetSlotList(CK_TRUE,
			                                  NULL_PTR,
			                                  &ulSlotCount);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			if (ulSlotCount == 0)
			{
				printf(" No Rutoken ECP is available!\n");
				break;
			}

			aSlots = (CK_SLOT_ID*)malloc(ulSlotCount * sizeof(CK_SLOT_ID));
			memset(aSlots,
			       0,
			       (ulSlotCount * sizeof(CK_SLOT_ID)));

			/**********************************************************************
			* 1.6 Получить список слотов c подключенными токенами                 *
			**********************************************************************/
			printf(" Getting list of connected slots");
			rv = pFunctionList->C_GetSlotList(CK_TRUE,
			                                  aSlots,
			                                  &ulSlotCount);
			if (rv != CKR_OK)
			{
				printf(" -> Failed %X\n", (int)rv);
				break;
			}
			printf(" -> OK\n");

			printf(" Slots available: 0x%8.8X\n", (int)ulSlotCount);

			/**********************************************************************
			* 1.7 Определить класс токена                                         *
			**********************************************************************/
			printf(" Determining token type");

			/**********************************************************************
			* Получить адрес функции запроса структуры с указателями на функции   *
			* расширения                                                          *
			**********************************************************************/
			pfGetFunctionListEx = (CK_C_EX_GetFunctionListExtended)GetProcAddress(hModule,
			                                                                      "C_EX_GetFunctionListExtended");
			if (pfGetFunctionListEx == NULL_PTR)
			{
				printf(" -> Failed\n");
				break;
			}

			/**********************************************************************
			* Получить структуру с указателями на функции расширения              *
			**********************************************************************/
			rv = pfGetFunctionListEx(&pFunctionListEx);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}

			memset(&tokenInfoEx,
			       0,
			       sizeof(CK_TOKEN_INFO_EXTENDED));
			tokenInfoEx.ulSizeofThisStructure = sizeof(CK_TOKEN_INFO_EXTENDED);

			/**********************************************************************
			* Получить расширенную информацию о подключенном токене               *
			**********************************************************************/
			rv = pFunctionListEx->C_EX_GetTokenInfoExtended(aSlots[0],
			                                                &tokenInfoEx);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}

			/**********************************************************************
			* Определить класс токена                                             *
			**********************************************************************/
			if (tokenInfoEx.ulTokenType == TOKEN_TYPE_RUTOKEN_PINPAD_FAMILY)
			{
				bIsRutokenPINPad = TRUE;
				printf(": Rutoken PINPad\n");
			}
			else if (tokenInfoEx.ulTokenType == TOKEN_TYPE_RUTOKEN_ECP)
			{
				bIsRutokenPINPad = FALSE;
				printf(": Rutoken ECP\n");
			}
			else if (tokenInfoEx.ulTokenType == TOKEN_TYPE_RUTOKEN_LITE)
			{
				bIsRutokenPINPad = FALSE;
				printf(": Rutoken Lite\n");
			}
			else if (tokenInfoEx.ulTokenType == TOKEN_TYPE_RUTOKEN)
			{
				bIsRutokenPINPad = FALSE;
				printf(": Rutoken / Rutoken S\n");
			}
			else
			{
				bIsRutokenPINPad = FALSE;
				printf(": undefined\n");
			}

			/**********************************************************************
			* 1.8 Открыть RW сессию в первом доступном слоте                      *
			**********************************************************************/
			printf(" Opening Session");
			rv = pFunctionList->C_OpenSession(aSlots[0],
			                                  CKF_SERIAL_SESSION | CKF_RW_SESSION,
			                                  NULL_PTR,
			                                  NULL_PTR,
			                                  &hSession);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/**********************************************************************
			* 1.9 Выполнить аутентификацию Пользователя                           *
			**********************************************************************/
			printf(" Logging in");
			rv = pFunctionList->C_Login(hSession,
			                            CKU_USER,
			                            USER_PIN,
			                            sizeof(USER_PIN));
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");
			break;
		}

		if ((rv != CKR_OK) || (ulSlotCount == 0))
		{
			printf("Initialization failed!\n");
			break;
		}
		else
			printf("Initialization has been completed successfully.\n");

		if (!bIsRutokenPINPad)
		{				
			printf("\n No Rutoken PINPad is available!\n");
			break;
		}

		/**********************************************************************
		* Шаг 2: Получить запись журнала                                      *
		**********************************************************************/
		/**********************************************************************
		* 2.1 Получить размер записи журнала                                  *
		**********************************************************************/
		printf("\nGetting journal size");
		rv = pFunctionListEx->C_EX_GetJournal(aSlots[0],
												NULL_PTR,
												&ulJournalSize);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");
 
		pJournal = (CK_BYTE*)malloc(ulJournalSize * sizeof(CK_BYTE));
		memset(pJournal, 0, (ulJournalSize * sizeof(CK_BYTE)));
			
		/**********************************************************************
		* 2.2 Получить запись журнала                                         *
		**********************************************************************/
		printf("Getting journal");
		rv = pFunctionListEx->C_EX_GetJournal(aSlots[0],
												pJournal,
												&ulJournalSize);
		if (rv != CKR_OK)
		{
			printf(" -> Failed %X\n", (int)rv);
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* 2.3 Распечатать содержимое журнала                                      *
		**********************************************************************/
		if (pJournal)
		{
			printJournal(pJournal, ulJournalSize);
			free(pJournal);
			pJournal = NULL_PTR;
		}
		else
		{
			printf("Journal is empty\n");
		}

		break;
	}

	/**********************************************************************
	* Шаг 3: Выполнить действия для завершения работы                     *
	*        с библиотекой PKCS#11                                        *
	**********************************************************************/
	printf("\nFinalizing... \n");
	rvTemp = CKR_OK;
	if (hSession)
	{
		/**********************************************************************
		* 3.1 Сбросить права доступа                                          *
		**********************************************************************/
		printf(" Logging out");
		rvTemp = pFunctionList->C_Logout(hSession);
		if ((rvTemp == CKR_OK) || (rvTemp == CKR_USER_NOT_LOGGED_IN))
			printf(" -> OK\n");
		else
			printf(" -> Failed\n");

		/**********************************************************************
		* 3.2 Закрыть все открытые сессии в слоте                             *
		**********************************************************************/
		printf(" C_CloseAllSession");
		rvTemp = pFunctionList->C_CloseAllSessions(aSlots[0]);
		if (rvTemp != CKR_OK)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");
		hSession = NULL_PTR;
	}

	if (pFunctionList)
	{
		/**********************************************************************
		* 3.3 Деинициализировать библиотеку                                   *
		**********************************************************************/
		printf(" Finalizing library");
		rvTemp = pFunctionList->C_Finalize(NULL_PTR);
		if (rvTemp != CKR_OK)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");
		pFunctionList = NULL_PTR;
	}

	if (hModule)
	{
		/**********************************************************************
		* 3.4 Выгрузить библиотеку из памяти                                  *
		**********************************************************************/
		printf(" Unloading");
		if (FreeLibrary(hModule) != TRUE)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");
		hModule = NULL_PTR;
	}

	if (rvTemp != CKR_OK)
		printf("Unloading failed!\n\n");
	else
		printf("Unloading has been completed successfully.\n\n");

	if (aSlots)
	{
		free(aSlots);
		aSlots = NULL_PTR;
	}

	if ((rv != CKR_OK) || (ulSlotCount == 0) || (ulSlotCount == 0) || (!bIsRutokenPINPad))
		printf("Some error occurred. Error code: 0x%8.8x. Press Enter to exit.\n", (int)rv);
	else if (rvTemp != CKR_OK)
		printf("Some error occurred. Error code: 0x%8.8x. Press Enter to exit.\n", (int)rvTemp);
	else
		printf("Test has been completed successfully. Press Enter to exit.\n");

	getchar();
	return rv != CKR_OK;
}
