/*************************************************************************
* Rutoken                                                                *
* Copyright (c) 2003-2015, CJSC Aktiv-Soft. All rights reserved.         *
* Подробная информация:  http://www.rutoken.ru                           *
* Загрузка драйверов:    http://www.rutoken.ru/hotline/download/drivers/ *
* Техническая поддержка: http://www.rutoken.ru/hotline/                  *
*------------------------------------------------------------------------*
* Пример работы с Рутокен PINPad при помощи библиотеки PKCS#11           *
* на языке C                                                             *
*------------------------------------------------------------------------*
* Использование команд работы с журналом:                                *
*  - установление соединения с Рутокен PINPad в первом доступном слоте;  *
*  - выполнение аутентификации c правами Пользователя;					 *
*  - получение содержимого журнала Рутокен PINPad;                       *
*  - генерация журнальной пары на Рутокен PINPad;                        *
*  - подпись содержимого журнала Рутокен PINPad журнальной парой;        *
*  - сброс прав доступа Пользователя на Рутокен PINPad и закрытие        *
*    соединения с Рутокен PINPad.                                        *
*------------------------------------------------------------------------*
* Пример самодостаточен                                                  *
*************************************************************************/

#include "Common.h"

/************************************************************************
* Шаблон для создания открытого ключа ГОСТ Р 34.10-2001 журнальной пары *
************************************************************************/
CK_ATTRIBUTE PubKeyJournal[] =
{
	{ CKA_CLASS, &ocPubKey, sizeof(ocPubKey)},                      // Объект открытого ключа ГОСТ Р 34.10-2001
	{ CKA_KEY_TYPE, &ktGOST34_10_2001, sizeof(ktGOST34_10_2001)},   // Тип ключа
	{ CKA_TOKEN, &bTrue, sizeof(bTrue)},                            // Ключ является объектом токена
	{ CKA_PRIVATE, &bFalse, sizeof(bFalse)},                        // Ключ доступен без авторизации на токене
	{ CKA_VENDOR_KEY_JOURNAL, &bTrue, sizeof(CK_BBOOL)},            // Ключ предназначен для работы с журналом 
};
 
/************************************************************************
* Шаблон для создания закрытого ключа ГОСТ Р 34.10-2001 журнальной пары *
************************************************************************/
CK_ATTRIBUTE PrivKeyJournal[] =
{
	{ CKA_CLASS, &ocPrivKey, sizeof(ocPrivKey)},                    // Объект закрытого ключа ГОСТ Р 34.10-2001
	{ CKA_KEY_TYPE, &ktGOST34_10_2001, sizeof(ktGOST34_10_2001)},   // Тип ключа
	{ CKA_TOKEN, &bTrue, sizeof(bTrue)},                            // Ключ является объектом токена
	{ CKA_PRIVATE, &bTrue, sizeof(bTrue)},                          // Ключ доступен только после авторизации на токене
	{ CKA_VENDOR_KEY_CONFIRM_OP, &bFalse, sizeof(bFalse) },         // Операция подписи НЕ требует подтверждения на PINPad
	{ CKA_VENDOR_KEY_JOURNAL, &bTrue, sizeof(CK_BBOOL)},            // Ключ предназначен для работы с журналом 
};

/* Шаблон для поиска открытого ключа журнальной пары */
CK_ATTRIBUTE PubKeyJournalTempl[] =
{
	{ CKA_VENDOR_KEY_JOURNAL, &bTrue, sizeof(CK_BBOOL)},
	{ CKA_CLASS, &ocPubKey, sizeof(ocPubKey)}
};

/* Шаблон для поиска закрытого ключа журнальной пары */
CK_ATTRIBUTE PrivKeyJournalTempl[] =
{
	{ CKA_VENDOR_KEY_JOURNAL, &bTrue, sizeof(CK_BBOOL)},
	{ CKA_CLASS, &ocPrivKey, sizeof(ocPrivKey)}
};
 
/************************************************************************
* Получить массив хэндлов объектов, соответствующих критериям поиска    *
************************************************************************/
BOOL FindObjects(IN CK_SESSION_HANDLE hSession,				  // Хэндл открытой сессии
				 IN CK_FUNCTION_LIST_PTR pFunctionList,       // Указатель на список функций PKCS#11, хранящийся в структуре CK_FUNCTION_LIST
				 IN CK_ATTRIBUTE_PTR pTemplate,               // Указатель на шаблон, в который помещены атрибуты для поиска
				 IN CK_ULONG ulCount,                         // Количество атрибутов в шаблоне поиска
				 OUT CK_OBJECT_HANDLE_PTR* pphObject,         // Указатель на массив хэндлов объектов, соответствующих критериям поиска
				 OUT CK_ULONG* pulObjectCount,                // Количество хэндлов в массиве
				 OUT CK_RV* prv                               // Код возврата. Могут быть возвращены только ошибки, определенные в PKCS#11
	)
{
	CK_RV rvTemp = CKR_OK;       // Вспомогательная переменная для хранения кода возврата
	*pulObjectCount = 0;
	*pphObject = NULL_PTR;

	while (TRUE)
	{
		/**********************************************************************
		* Инициализировать операцию поиска                                    *
		**********************************************************************/
		printf(" C_FindObjectsInit");
		*prv = pFunctionList->C_FindObjectsInit(hSession,
		                                        pTemplate,
		                                        ulCount);
		if (*prv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* Найти все объекты, соответствующие критериям поиска                 *
		**********************************************************************/
		printf(" C_FindObjects");

		// Считаем, что максимальное количество объектов не превышает 100
		*pphObject = (CK_OBJECT_HANDLE*)malloc(100 * sizeof(CK_OBJECT_HANDLE));
		memset(*pphObject,
		       0,
		       (100 * sizeof(CK_OBJECT_HANDLE)));

		*prv = pFunctionList->C_FindObjects(hSession,
		                                    *pphObject,
		                                    100,
		                                    pulObjectCount);
		if (*prv != CKR_OK)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");
		break;
	}

	/**********************************************************************
	* Деинициализировать операцию поиска                                  *
	**********************************************************************/
	printf(" C_FindObjectsFinal");
	rvTemp = pFunctionList->C_FindObjectsFinal(hSession);
	if (rvTemp != CKR_OK)
		printf(" -> Failed\n");
	else
		printf(" -> OK\n");

	if (*prv == CKR_OK)
		printf("Search has been completed.\n"
		       "Objects found: %d \n",
		       (int)*pulObjectCount);
	else
	{
		printf("Search failed!\n");
		if (pphObject)
		{
			free(pphObject);
			*pphObject = NULL_PTR;
			*pulObjectCount = 0;
		}
	}

	return *prv == CKR_OK;
}

/**********************************************************************
* Сформировать хеш от исходных данных                                 *
**********************************************************************/
BOOL HashData(IN CK_SESSION_HANDLE hSession,				// Хэндл открытой сессии
			  IN CK_FUNCTION_LIST_PTR pFunctionList,        // Указатель на список функций PKCS#11, хранящийся в структуре CK_FUNCTION_LIST
			  IN CK_MECHANISM_PTR pHashMechanism,           // Указатель на механизм хеширования данных
			  IN CK_BYTE_PTR pbtDataToHash,                 // Указатель на буфер с данными для хеширования
			  IN CK_ULONG ulDataToHashSize,                 // Размер буфера с данными для хеширования в байтах
			  OUT CK_BYTE_PTR* ppbtHash,                    // Указатель на буфер, в который будут помещены хешированные данные
			  OUT CK_ULONG* pulHashSize,                    // Размер буфера, в который будут помещены хешированные данные, в байтах
			  OUT CK_RV* prv                                // Код возврата. Могут быть возвращены только ошибки, определенные в PKCS#11
	)
{
	DWORD i = 0;               // Вспомогательная переменная. Счетчик цикла
	*ppbtHash = NULL_PTR;
	*pulHashSize = 0;

	printf(" Hashing data...\n");
	while (TRUE)
	{
		/**********************************************************************
		* Инициализировать операцию хеширования                               *
		**********************************************************************/
		printf("  C_DigestInit");
		*prv = pFunctionList->C_DigestInit(hSession,
										   pHashMechanism);
		if (*prv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* Определить размер хешированных данных                               *
		**********************************************************************/
		printf("  C_Digest step 1");
		*prv = pFunctionList->C_Digest(hSession,
									   pbtDataToHash,
									   ulDataToHashSize,
									   *ppbtHash,
									   pulHashSize);
		if (*prv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		*ppbtHash = (CK_BYTE*)malloc(*pulHashSize * sizeof(CK_BYTE));
		memset(*ppbtHash,
			   0,
			   (*pulHashSize * sizeof(CK_BYTE)));

		/**********************************************************************
		* Сформировать хеш от исходных данных                                 *
		**********************************************************************/
		printf("  C_Digest step 2");
		*prv = pFunctionList->C_Digest(hSession,
										pbtDataToHash,
										ulDataToHashSize,
										*ppbtHash,
										pulHashSize);
		if (*prv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/************************************************************************
		* Распечатать буфер, содержащий хешированные данные                     *
		************************************************************************/
		printf(" Hashed buffer is: \n");

		for (i = 0;
			 i < *pulHashSize;
			 i++)
		{
			printf(" %02X", (*ppbtHash)[i]);
			if ((i + 1) % 8 == 0)
				printf("\n");
		}
		break;
	}

	if ((*prv != CKR_OK) && (*ppbtHash))
	{
		free(*ppbtHash);
		*ppbtHash = NULL_PTR;
		*pulHashSize = 0;
	}

	if (*prv == CKR_OK)
		printf(" Hashing has been completed.\n");
	else
		printf("\n\n Hashing failed!\n\n");

	return *prv == CKR_OK;
}


/************************************************************************
* main()                                                                *
************************************************************************/
int main(void)
{
	HMODULE hModule = NULL_PTR;                          // Хэндл загруженной библиотеки PKCS#11
	CK_SESSION_HANDLE hSession = NULL_PTR;               // Хэндл открытой сессии

	CK_FUNCTION_LIST_PTR pFunctionList = NULL_PTR;       // Указатель на список функций PKCS#11, хранящийся в структуре CK_FUNCTION_LIST
	CK_C_GetFunctionList pfGetFunctionList = NULL_PTR;   // Указатель на функцию C_GetFunctionList

	CK_C_EX_GetFunctionListExtended pfGetFunctionListEx = NULL_PTR; // Указатель на функцию C_EX_GetFunctionListExtended
	CK_FUNCTION_LIST_EXTENDED_PTR pFunctionListEx = NULL_PTR;       // Указатель на список функций расширения PKCS#11, хранящийся в структуре CK_FUNCTION_LIST_EXTENDED
	CK_TOKEN_INFO_EXTENDED tokenInfoEx;                             // Структура данных типа CK_TOKEN_INFO_EXTENDED с информацией о токене

	CK_SLOT_ID_PTR aSlots = NULL_PTR;                    // Указатель на массив идентификаторов слотов
	CK_ULONG ulSlotCount = 0;                            // Количество идентификаторов слотов в массиве

	CK_OBJECT_HANDLE hPubKey = NULL_PTR;				 // Хэндл открытого ключа журнальной пары
	CK_OBJECT_HANDLE hPrivKey = NULL_PTR;				 // Хэндл закрытого ключа журнальной пары

	CK_OBJECT_HANDLE_PTR phObject = NULL_PTR;            // Указатель на массив хэндлов объектов, соответствующих критериям поиска
	CK_ULONG ulObjectCount = 0;                          // Количество хэндлов объектов в массиве

	CK_BYTE_PTR pJournal = NULL_PTR;					 // Указатель на значение журнала
	CK_ULONG ulJournalSize = 0;							 // Размер журнала
	
	CK_BYTE_PTR pbtSignature = NULL_PTR;                 // Указатель на буфер, содержащий подпись для исходных данных
	CK_ULONG ulSignatureSize = 0;                        // Размер буфера, содержащего подпись для исходных данных, в байтах

	CK_BYTE_PTR pbHash = NULL_PTR;                       // Указатель на временный буфер для хешированных данных
	CK_ULONG ulHashSize = 0;                             // Размер временного буфера в байтах

	CK_BBOOL bIsRutokenPINPad = FALSE;                   // Вспомогательная переменная для хранения признака типа токена
	CK_RV rv = CKR_OK;                                   // Вспомогательная переменная для хранения кода возврата
	CK_RV rvTemp = CKR_OK;                               // Вспомогательная переменная для хранения кода возврата

	DWORD i = 0;                                         // Вспомогательная переменная-счетчик в циклах
	
	CK_ATTRIBUTE attrPublicValue = { CKA_VALUE, NULL_PTR, 0 }; // Шаблон для получения атрибутов открытого ключа

	while (TRUE)
	{
		/**********************************************************************
		* Шаг 1: Выполнить действия для начала работы с библиотекой PKCS#11.  *
		**********************************************************************/
		printf("Initialization module...\n");
		while (TRUE)
		{
			/**********************************************************************
			* 1.1 Загрузить библиотеку                                            *
			**********************************************************************/
			printf("Loading library %s", PKCS11ECP_LIBRARY_NAME);
			hModule = LoadLibrary(PKCS11ECP_LIBRARY_NAME);
			if (hModule == NULL_PTR)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/************************************************************************
			* 1.2 Получить адрес функции запроса структуры с указателями на функции *
			************************************************************************/
			printf(" Getting GetFunctionList function");
			pfGetFunctionList = (CK_C_GetFunctionList)GetProcAddress(hModule,
			                                                         "C_GetFunctionList");
			if (pfGetFunctionList == NULL_PTR)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/**********************************************************************
			* 1.3 Получить структуру с указателями на функции                     *
			**********************************************************************/
			printf(" Getting function list");
			rv = pfGetFunctionList(&pFunctionList);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/**********************************************************************
			* 1.4 Инициализировать библиотеку                                     *
			**********************************************************************/
			printf(" Initializing library");
			rv = pFunctionList->C_Initialize(NULL_PTR);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/**********************************************************************
			* 1.5 Получить количество слотов c подключенными токенами             *
			**********************************************************************/
			printf(" Getting number of connected slots");
			rv = pFunctionList->C_GetSlotList(CK_TRUE,
			                                  NULL_PTR,
			                                  &ulSlotCount);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			if (ulSlotCount == 0)
			{
				printf(" No Rutoken ECP is available!\n");
				break;
			}

			aSlots = (CK_SLOT_ID*)malloc(ulSlotCount * sizeof(CK_SLOT_ID));
			memset(aSlots,
			       0,
			       (ulSlotCount * sizeof(CK_SLOT_ID)));

			/**********************************************************************
			* 1.6 Получить список слотов c подключенными токенами                 *
			**********************************************************************/
			printf(" Getting list of connected slots");
			rv = pFunctionList->C_GetSlotList(CK_TRUE,
			                                  aSlots,
			                                  &ulSlotCount);
			if (rv != CKR_OK)
			{
				printf(" -> Failed %X\n", (int)rv);
				break;
			}
			printf(" -> OK\n");

			printf(" Slots available: 0x%8.8X\n", (int)ulSlotCount);

			/**********************************************************************
			* 1.7 Определить класс токена                                         *
			**********************************************************************/
			printf(" Determining token type");

			/**********************************************************************
			* Получить адрес функции запроса структуры с указателями на функции   *
			* расширения                                                          *
			**********************************************************************/
			pfGetFunctionListEx = (CK_C_EX_GetFunctionListExtended)GetProcAddress(hModule,
			                                                                      "C_EX_GetFunctionListExtended");
			if (pfGetFunctionListEx == NULL_PTR)
			{
				printf(" -> Failed\n");
				break;
			}

			/**********************************************************************
			* Получить структуру с указателями на функции расширения              *
			**********************************************************************/
			rv = pfGetFunctionListEx(&pFunctionListEx);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}

			memset(&tokenInfoEx,
			       0,
			       sizeof(CK_TOKEN_INFO_EXTENDED));
			tokenInfoEx.ulSizeofThisStructure = sizeof(CK_TOKEN_INFO_EXTENDED);

			/**********************************************************************
			* Получить расширенную информацию о подключенном токене               *
			**********************************************************************/
			rv = pFunctionListEx->C_EX_GetTokenInfoExtended(aSlots[0],
			                                                &tokenInfoEx);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}

			/**********************************************************************
			* Определить класс токена                                             *
			**********************************************************************/
			if (tokenInfoEx.ulTokenType == TOKEN_TYPE_RUTOKEN_PINPAD_FAMILY)
			{
				bIsRutokenPINPad = TRUE;
				printf(": Rutoken PINPad\n");
			}
			else if (tokenInfoEx.ulTokenType == TOKEN_TYPE_RUTOKEN_ECP)
			{
				bIsRutokenPINPad = FALSE;
				printf(": Rutoken ECP\n");
			}
			else if (tokenInfoEx.ulTokenType == TOKEN_TYPE_RUTOKEN_LITE)
			{
				bIsRutokenPINPad = FALSE;
				printf(": Rutoken Lite\n");
			}
			else if (tokenInfoEx.ulTokenType == TOKEN_TYPE_RUTOKEN)
			{
				bIsRutokenPINPad = FALSE;
				printf(": Rutoken / Rutoken S\n");
			}
			else
			{
				bIsRutokenPINPad = FALSE;
				printf(": undefined\n");
			}

			/**********************************************************************
			* 1.8 Открыть RW сессию в первом доступном слоте                      *
			**********************************************************************/
			printf(" Opening Session");
			rv = pFunctionList->C_OpenSession(aSlots[0],
			                                  CKF_SERIAL_SESSION | CKF_RW_SESSION,
			                                  NULL_PTR,
			                                  NULL_PTR,
			                                  &hSession);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/**********************************************************************
			* 1.9 Выполнить аутентификацию Пользователя                           *
			**********************************************************************/
			printf(" Logging in");
			rv = pFunctionList->C_Login(hSession,
			                            CKU_USER,
			                            USER_PIN,
			                            sizeof(USER_PIN));
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");
			break;
		}

		if ((rv != CKR_OK) || (ulSlotCount == 0))
		{
			printf("Initialization failed!\n");
			break;
		}
		else
			printf("Initialization has been completed successfully.\n");

		if (!bIsRutokenPINPad)
		{				
			printf("\n No Rutoken PINPad is available!\n");
			break;
		}

		/**********************************************************************
		* Шаг 2: Получить запись журнала                                      *
		**********************************************************************/
		/**********************************************************************
		* 2.1 Получить размер записи журнала                                  *
		**********************************************************************/
		printf("\nGetting journal size");
		rv = pFunctionListEx->C_EX_GetJournal(aSlots[0],
												NULL_PTR,
												&ulJournalSize);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");
 
		pJournal = (CK_BYTE*)malloc(ulJournalSize * sizeof(CK_BYTE));
		memset(pJournal, 0, (ulJournalSize * sizeof(CK_BYTE)));
			
		/**********************************************************************
		* 2.2 Получить запись журнала                                         *
		**********************************************************************/
		printf("Getting journal");
		rv = pFunctionListEx->C_EX_GetJournal(aSlots[0],
												pJournal,
												&ulJournalSize);
		if (rv != CKR_OK)
		{
			printf(" -> Failed %X\n", (int)rv);
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* 2.3 Распечатать содержимое журнала                                      *
		**********************************************************************/
		printf("Journal buffer is: \n");
		for (i = 0;
			i < ulJournalSize;
			i++)
		{
			printf("%02X ", pJournal[i]);
			if ((i + 1) % 8 == 0)
				printf("\n");
		}	
			
		/************************************************************************
		* Шаг 3: Получить хэндл закрытого ключа журнальной пары                 *
		************************************************************************/
		printf("\n\nGetting journal private key...\n");
		FindObjects(hSession,
					pFunctionList,
					PrivKeyJournalTempl,
					arraysize(PrivKeyJournalTempl),
					&phObject,
					&ulObjectCount,
					&rv);
		if (rv != CKR_OK)
			break;

		if (ulObjectCount == 0)
		{
			printf("\nNo journal key found!\n");

			/**********************************************************************
			* Шаг 4: Сгенерировать журнальную ключевую пару                       *
			**********************************************************************/
			printf("\nGenerating key pair");
			rv = pFunctionList->C_GenerateKeyPair(hSession,                     // Хэндл открытой сессии
													&ckmGOST34_10_2001KeyGenMech, // Используемый механизм генерации ключевой пары ГОСТ Р 34.10-2001
													PubKeyJournal,                // Шаблон открытого ключа журнальной пары
													arraysize(PubKeyJournal),     // Размер шаблона открытого ключа
													PrivKeyJournal,               // Шаблон закрытого ключа журнальной пары
													arraysize(PrivKeyJournal),    // Размер шаблона закрытого ключа
													&hPubKey,                     // Хэндл открытого ключа журнальной пары
													&hPrivKey);                   // Хэндл закрытого ключа журнальной пары
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			printf("Generation journal key pair has been completed successfully.\n");

			printf("\nGetting journal private key...\n");
			FindObjects(hSession,
						pFunctionList,
						PrivKeyJournalTempl,
						arraysize(PrivKeyJournalTempl),
						&phObject,
						&ulObjectCount,
						&rv);
		}

		if (rv != CKR_OK)
			break;

		/**********************************************************************
		* Шаг 5: Подписать содержимое журнала                                 *
		**********************************************************************/			
		printf("\nSigning journal...\n");
			
		/**********************************************************************
		* 5.1 Сформировать хеш от содержимого журнала                         *
		**********************************************************************/
		HashData(hSession,
			        pFunctionList,
			        &ckmGOST34_11_94Mech,
					pJournal,
					ulJournalSize,
			        &pbHash,
			        &ulHashSize,
			        &rv);
		if (rv != CKR_OK)
			break;

		/**********************************************************************
		* 5.2 Инициализировать операцию подписи журнала                       *
		**********************************************************************/
		printf(" C_EX_SignInvisibleInit");
		rv = pFunctionListEx->C_EX_SignInvisibleInit(hSession,
														&ckmGOST_34_10_2001SigVerMech,
										                phObject[0]);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/**********************************************************************
		* 5.3 Определить размер подписанных данных                            *
		**********************************************************************/
		printf(" C_EX_SignInvisible step 1");
		rv = pFunctionListEx->C_EX_SignInvisible(hSession,
												    pbHash,
												    ulHashSize,
												    pbtSignature,
												    &ulSignatureSize);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		pbtSignature = (CK_BYTE*)malloc(ulSignatureSize);
		memset(pbtSignature,
				0,
				ulSignatureSize * sizeof(CK_BYTE));

		/**********************************************************************
		* 5.4 Подписать содержимое журнала                                    *
		**********************************************************************/
		printf(" C_EX_SignInvisible step 2");
		rv = pFunctionListEx->C_EX_SignInvisible(hSession,
												    pbHash,
												    ulHashSize,
												    pbtSignature,
												    &ulSignatureSize);
		if (rv != CKR_OK)
		{
			printf(" -> Failed\n");
			break;
		}
		printf(" -> OK\n");

		/************************************************************************
		* 5.5 Распечатать буфер, содержащий подпись                             *
		************************************************************************/
		printf(" Signature buffer is: \n ");
		for (i = 0;
			i < ulSignatureSize;
			i++)
		{
			printf("%02X ", pbtSignature[i]);
			if ((i + 1) % 8 == 0)
				printf("\n ");
		}

		if ((rv != CKR_OK) || (ulObjectCount == 0))
		{
			printf("\nSigning operation failed!\n");
			break;
		}
		else
		printf("Journal has been signed successfully.\n");

		if (pbHash)
		{
			free(pbHash);
			pbHash = NULL_PTR;
			ulHashSize = 0;
		}	
			
		if (phObject)
		{
			free(phObject);
			phObject = NULL_PTR;
			ulObjectCount = 0;
		}

		/**********************************************************************
		* Шаг 6: Выполнить проверку подписи содержимого журнала               *
		*        по алгоритму ГОСТ Р 34.10-2001                               *
		**********************************************************************/
		printf("\nVerifying signature...\n");

		while (TRUE)
		{
			/************************************************************************
			* 6.1 Получить массив хэндлов открытых ключей                           *
			************************************************************************/
			printf(" Getting key to verify...\n");
						FindObjects(hSession,
						pFunctionList,
						PubKeyJournalTempl,
						arraysize(PubKeyJournalTempl),
						&phObject,
						&ulObjectCount,
						&rv);
			if (rv != CKR_OK)
				break;

			if (ulObjectCount == 0)
			{
				printf("\nNo verification key found!\n");
				break;
			}

			/**********************************************************************
			* 6.2 Сформировать хеш от исходных данных                             *
			**********************************************************************/
			free(pbHash);
			pbHash = NULL_PTR;
			HashData(hSession,
						pFunctionList,
						&ckmGOST34_11_94Mech,
						pJournal,
						ulJournalSize,
						&pbHash,
						&ulHashSize,
						&rv);
			if (rv != CKR_OK)
				break;

			/**********************************************************************
			* 6.3 Инициализировать операцию проверки подписи                      *
			**********************************************************************/
			printf(" C_VerifyInit");
			rv = pFunctionList->C_VerifyInit(hSession,
												&ckmGOST_34_10_2001SigVerMech,
												phObject[0]);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");

			/**********************************************************************
			* 6.4 Проверить подпись для исходных данных                           *
			**********************************************************************/
			printf(" C_Verify");
			rv = pFunctionList->C_Verify(hSession,
											pbHash,
											ulHashSize,
											pbtSignature,
											ulSignatureSize);
			if (rv != CKR_OK)
			{
				printf(" -> Failed\n");
				break;
			}
			printf(" -> OK\n");
			break;
		}

		if ((rv != CKR_OK) || (ulObjectCount == 0))
		{
			printf("\nVerifying failed!\n\n");
			break;
		}
		else
			printf("\nVerifying has been completed successfully.\n");
		
		if (pbHash)
		{
			free(pbHash);
			pbHash = NULL_PTR;
		}	
		
		if (pbtSignature)
		{
			free(pbtSignature);
			pbtSignature = NULL_PTR;
		}

		if (pJournal)
		{
			free(pJournal);
			pJournal = NULL_PTR;
		}

		if (phObject)
		{
			free(phObject);
			phObject = NULL_PTR;
		}

		break;
	}

	/**********************************************************************
	* Шаг 6: Выполнить действия для завершения работы                     *
	*        с библиотекой PKCS#11                                        *
	**********************************************************************/
	printf("\nFinalizing... \n");
	rvTemp = CKR_OK;
	if (hSession)
	{
		/**********************************************************************
		* 6.1 Сбросить права доступа                                          *
		**********************************************************************/
		printf(" Logging out");
		rvTemp = pFunctionList->C_Logout(hSession);
		if ((rvTemp == CKR_OK) || (rvTemp == CKR_USER_NOT_LOGGED_IN))
			printf(" -> OK\n");
		else
			printf(" -> Failed\n");

		/**********************************************************************
		* 6.2 Закрыть все открытые сессии в слоте                             *
		**********************************************************************/
		printf(" C_CloseAllSession");
		rvTemp = pFunctionList->C_CloseAllSessions(aSlots[0]);
		if (rvTemp != CKR_OK)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");
		hSession = NULL_PTR;
	}

	if (pFunctionList)
	{
		/**********************************************************************
		* 6.3 Деинициализировать библиотеку                                   *
		**********************************************************************/
		printf(" Finalizing library");
		rvTemp = pFunctionList->C_Finalize(NULL_PTR);
		if (rvTemp != CKR_OK)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");
		pFunctionList = NULL_PTR;
	}

	if (hModule)
	{
		/**********************************************************************
		* 6.4 Выгрузить библиотеку из памяти                                  *
		**********************************************************************/
		printf(" Unloading");
		if (FreeLibrary(hModule) != TRUE)
			printf(" -> Failed\n");
		else
			printf(" -> OK\n");
		hModule = NULL_PTR;
	}

	if (rvTemp != CKR_OK)
		printf("Unloading failed!\n\n");
	else
		printf("Unloading has been completed successfully.\n\n");

	if (aSlots)
	{
		free(aSlots);
		aSlots = NULL_PTR;
	}

	if ((rv != CKR_OK) || (ulSlotCount == 0) || (ulSlotCount == 0) || (!bIsRutokenPINPad))
		printf("Some error occurred. Error code: 0x%8.8x. Press Enter to exit.\n", (int)rv);
	else if (rvTemp != CKR_OK)
		printf("Some error occurred. Error code: 0x%8.8x. Press Enter to exit.\n", (int)rvTemp);
	else
		printf("Test has been completed successfully. Press Enter to exit.\n");

	getchar();
	return rv != CKR_OK;
}
